/*
RUNESPEL
by blue_bear_94
Programming Contest against Nspire_Lover
*/
#ifndef USE_TI89
#define USE_TI89            // produce all types of files (compatible for all calcs)
#define USE_TI92PLUS
#define USE_V200
#endif
#include <tigcclib.h> //all TIGCC header files (here, graph.h, gray.h, dialogs.h, stdio.h, stdlib.h, sprites.h, kbd.h)
//constant definitions
#define ROWS 10//rows in the grid
#define COLS 20//columns in the grid
#define MAXTIME 2100//starting time (in 1/20 of seconds) in timed mode
#define MAXMOVES 30//number of moves in puzzle mode

const SCR_RECT mainRect=(SCR_RECT){{0,0,119,99}};
const SCR_RECT wholeScn=(SCR_RECT){{0,0,159,99}};
enum Directions {RIGHT,DOWN};//Enum for match directions.
enum RuneNames {R=1,U,N,E,S,P,L};//Enum for rune names.
enum PlayModes {TIMED,PUZZLE};//Enum for playing modes.
enum GrayShades {WHITE,LGRAY,DGRAY,BLACK};//Enum describing 4-level grayscale.
const char* errors[]={"DMA Failure","Could not allocate memory"};//Error messages.
int sf=0;//Number of matches.
//9*6
char rpics[7][7]={{0xC0,0xB0,0x88,0xB0,0xC0,0xB0,0x88},{0x80,0xE0,0x98,0x88,0x88,0x88,0x88},{0x08,0x88,0xC8,0xA8,0x98,0x88,0x80},{0x88,0xD8,0xA8,0x88,0x88,0x88,0x88},{0x80,0x88,0x98,0xA8,0xC8,0x88,0x08},
{0x10,0x30,0x50,0x90,0x10,0x10,0x10},{0x20,0x20,0x60,0xA8,0x30,0x20,0x20}};//Rune images.
char rgray[7]={3,1,2,3,1,2,3};//Rune colors.
long hsTimed[11]={10000,9000,8000,7000,6000,5000,4000,3000,2000,1000,0};
long hsPuzzle[11]={10000,9000,8000,7000,6000,5000,4000,3000,2000,1000,0};
CALLBACK long cmp(const void* a,const void* b)
{
	return -fcmp(*(const long*)a,*(const long*)b);
}
typedef struct //coordinates
{
	unsigned char row:4,col:5;
} boardPos;
typedef struct //match
{
	boardPos head;
	unsigned char dir:1,len:3;
} Match;
void clrField(void)
{
	GraySetAMSPlane(0);
	ScrRectFill(&mainRect,&wholeScn,A_REVERSE);
	GraySetAMSPlane(1);
	ScrRectFill(&mainRect,&wholeScn,A_REVERSE);
}
inline void throwErr(int title,int text)//inline function to display error dialog box
{
	DlgMessage(errors[title],errors[text],BT_NONE,BT_CANCEL);
}
void ClrScrAll(void)
{
	GraySetAMSPlane(0);
	ClrScr();
	GraySetAMSPlane(1);
	ClrScr();
}
void drawRune(short x,short y,short num)
{
	if (num)
	{
		if (rgray[num-1]&1) Sprite8(x,y,7,rpics[num-1],GrayGetPlane(LIGHT_PLANE),SPRT_XOR);
		if (rgray[num-1]&2) Sprite8(x,y,7,rpics[num-1],GrayGetPlane(DARK_PLANE),SPRT_XOR);
	}
}
void drawField(char** field)//Draws the runefield.
{
	int i,j;
	for(i=0;i<ROWS;i++)
	{
		for(j=0;j<COLS;j++)
		{
			drawRune(6*j,9*i,field[i][j]);
		}
	}
}
Match* findMatches(char** field) // returns an array of Match structure. Must free DMA array yourself.
{
	Match* x=NULL;
	int nom=1;
	if (!(x=(Match*)calloc(1,sizeof(Match))))
	{
		return x;
	}
	int i,j;
	//horiz
	for (i=0;i<ROWS;i++)
	{
		for (j=0;j<COLS-2;j++)
		{
			if((field[i][j]==field[i][j+1])&&(field[i][j]==field[i][j+2]))
			{
				if (!(x=(Match*)realloc(x,sizeof(Match)*(nom+1))))
				{
					return x;
				}
				nom++;
				Match* y=&(x[nom-2]);
				(y->head).row=i;
				(y->head).col=j;
				y->dir=RIGHT;
				y->len=3+(j<17?field[i][j]==field[i][j+3]:0)+(j<16?field[i][j]==field[i][j+4]:0)+(j<15?field[i][j]==field[i][j+5]:0);
				j+=y->len-1;
			}
		}
	}
	//vert
	for (i=0;i<COLS;i++)
	{
		for (j=0;j<ROWS-2;j++)
		{
			if((field[j][i]==field[j+1][i])&&(field[j][i]==field[j+2][i]))
			{
				if (!(x=(Match*)realloc(x,sizeof(Match)*(nom+1))))
				{
					return x;
				}
				nom++;
				Match* y=&(x[nom-2]);
				(y->head).row=j;
				(y->head).col=i;
				y->dir=DOWN;
				y->len=3+(j<7?field[j][i]==field[j+3][i]:0)+(j<6?field[j][i]==field[j+4][i]:0)+(j<5?field[j][i]==field[j+5][i]:0);
				j+=y->len-1;
			}
		}
	}
	sf=nom-1;
	return x;
}
long newScore(long currentScore,Match *x,short combo) //returns updated score
{
	long score=currentScore;
	int i;
	for(i=0;i<sf;i++)
	{
		int y=x[i].len;
		score+=(100*(y==3)+250*(y==4)+500*(y==5)+1000*(y>5))*(0.1*combo+1);
	}
	return score;
}
void swap(char** field,int r1,int c1,int r2, int c2)//swaps 2 elements in an array
{
	char tmp1=field[r1][c1];
	field[r1][c1]=field[r2][c2];
	field[r2][c2]=tmp1;
}
void deleteMatches(Match *x,char** field) //sets runes in matches to 0
{
	int i;
	for(i=0;i<sf;i++)
	{
		Match y=x[i];
		int u,v;
		u=y.head.row;
		v=y.head.col;
		int j;
		for(j=0;j<y.len;j++)
		{
			field[u+j*(y.dir==DOWN)][v+j*(y.dir==RIGHT)]=0;
		}
	}
}
int sumCol(char** field,int col) //returns the number of filled cells in a given column
{
	int i,num=0;
	for(i=0;i<ROWS;i++)
	{
		num+=!!(field[i][col]);
	}
	return num;
}
void shiftField(char** field)//shifts floating runes down and fills empty spaces with runes
{
	int col;
	for(col=0;col<COLS;col++)//for each of the cols...
	{
		int bottomEmpty=9;
		//while there are empty cells in the column
		while (bottomEmpty!=-1)
		{
			//find lowermost empty cell
			bottomEmpty=ROWS-1;
			while (bottomEmpty>=0?field[bottomEmpty][col]:0)//if it's -1 there's no need to continue
			{
				bottomEmpty--;
			}
			//holds -1 if no such cell. If bottomEmpty is -1 or 0, then there is no need to shift.
			if (bottomEmpty>0)
			{
				//shift above cells down
				int i;
				for(i=bottomEmpty-1;i>-1;i--)
				{
					field[i+1][col]=field[i][col];
				}
			}
		//fill topmost cell
		if (bottomEmpty!=-1) field[0][col]=1+random(7);
		}
	}
}
int trySwap(char** field,int r1,int c1,int r2, int c2,long* sptr,int combo)//swaps 2 elements in an array, and if there are no matches, swaps back. If there are, then does the usual... (0 if no match, 1 if match, 2 if add'l matches)
{
	//add graphic effects!
	clrField();
	swap(field,r1,c1,r2,c2);
	Match* Z=NULL;
	Z=findMatches(field);
	if (!sf) {swap(field,r1,c1,r2,c2); free(Z); drawField(field); return 0;}//swaps back if no matches are found
	else
	{
		deleteMatches(Z,field);
		drawField(field);
		OSRegisterTimer(USER1_TIMER,5);
		shiftField(field);
		*sptr=newScore(*sptr,Z,combo);
		while (!OSTimerExpired(USER1_TIMER));
		clrField();
		drawField(field);
		OSFreeTimer(USER1_TIMER);
		free(Z);
		Z=findMatches(field);
		free(Z);
		return 1+!!sf;
	}
	
}


inline void drawCurs(int row,int col)//normal cursor
{
	GraySetAMSPlane(LIGHT_PLANE);
	DrawLine(6*col,9*row+8,6*col+5,9*row+8,A_XOR);
}
inline void drawCursSwap(int row,int col)//swap cursor
{
	GraySetAMSPlane(DARK_PLANE);
	DrawLine(6*col,9*row+8,6*col+5,9*row+8,A_XOR);
}
void drawRBlk(int x,int y, int num)
{
	Sprite8(x,y,7,rpics[num-1],GrayGetPlane(LIGHT_PLANE),SPRT_XOR);
	Sprite8(x,y,7,rpics[num-1],GrayGetPlane(DARK_PLANE),SPRT_XOR);
}
void DrawInt(int x,int y,long num,int attr)
{
	char buff[11];
	sprintf(buff,"%ld   ",num);
	DrawStr(x,y,buff,attr);
}
void drwRtSec(int mode,long score,int time,int moves)
{
	int plane;
	for(plane=0;plane<2;plane++)
	{
		//set plane
		GraySetAMSPlane(plane);
		//borders
		DrawLine(120,0,120,99,A_NORMAL);
		DrawLine(120,17,159,17,A_NORMAL);
		//draw "RUNESPEL"
		drawRBlk(122,1,R);
		drawRBlk(128,1,U);
		drawRBlk(134,1,N);
		drawRBlk(140,1,E);
		drawRBlk(128,9,S);
		drawRBlk(134,9,P);
		drawRBlk(140,9,E);
		drawRBlk(146,9,L);
		//details...
		FontSetSys(F_4x6);//small font
		char buff[11];
		DrawStr(122,19,"SCORE",A_NORMAL);
		sprintf(buff,"%08ld",score);
		DrawStr(122,25,buff,A_REPLACE);
		DrawStr(122,45,"HIGH",A_REPLACE);
		sprintf(buff,"%08ld",mode?hsPuzzle[0]:hsTimed[0]);
		DrawStr(122,51,buff,A_REPLACE);
		if (mode==TIMED)
		{
			DrawStr(122,31,"TIME",A_NORMAL);
			FontSetSys(F_6x8);
			DrawInt(122,37,time/20,A_REPLACE);
		}
		else
		{
			DrawStr(122,31,"MOVES",A_NORMAL);
			FontSetSys(F_6x8);
			DrawInt(122,37,moves,A_REPLACE);
		}
	}
}
void startSession(int mode)//Starts a game.
{
	ClrScrAll();
	//initialize vars
	int moves=MAXMOVES;
	long score=0;
	char **field = NULL;
	int i,j;
	field = calloc (ROWS, sizeof (*field));
	for (i = 0; i < 10; i++)
  field[i] = calloc (20, sizeof (char));

	//char field[10][20];
	boardPos curs;
	curs.row=0;
	curs.col=0;
	int swapMode=0;
	if (mode==TIMED) OSRegisterTimer(USER_TIMER,MAXTIME);
	int key;
	for(i=0;i<ROWS;i++)
	{
		for(j=0;j<COLS;j++)
		{
			field[i][j]=random(7)+1;
		}
	}
	drawField(field);
	drawCurs(curs.row,curs.col);
	while (mode==TIMED ? !OSTimerExpired(USER_TIMER) : moves)
	{
		//all the actions
		//get key
		key=kbhit()?ngetchx():0;
		//move cursor
		if (!swapMode)
		{
			drawCurs(curs.row,curs.col);
			if (key==KEY_UP && curs.row!=0) curs.row--;
			if (key==KEY_DOWN && curs.row!=ROWS-1) curs.row++;
			if (key==KEY_LEFT && curs.col!=0) curs.col--;
			if (key==KEY_RIGHT && curs.col!=COLS-1) curs.col++;
			drawCurs(curs.row,curs.col);
		}
		else //means swap
		{
			int p= -1,q= -1;
			if (key==KEY_LEFT && curs.col!=0) {p=curs.row; q=curs.col-1;}
			if (key==KEY_RIGHT && curs.col!=COLS-1) {p=curs.row; q=curs.col+1;}
			if (key==KEY_UP && curs.row!=0) {p=curs.row-1; q=curs.col;}
			if (key==KEY_DOWN && curs.row!=ROWS-1) {p=curs.row+1; q=curs.col;}
			if (p>=0 && p<ROWS && q>=0 && q<COLS)
			{
				drawCursSwap(curs.row,curs.col);
				moves-=!!trySwap(field,curs.row,curs.col,p,q,&score,0);
				curs.row=p; curs.col=q;
				int combo=1;
				while (trySwap(field,curs.row,curs.col,p,q,&score,combo)==2) {combo++;};
				drawCurs(curs.row,curs.col);
				swapMode=0;
			}
			
		}
		if (key==KEY_ENTER||key==KEY_F1)
		{
			swapMode=!swapMode;
			if (swapMode) //light to dark
			{
				drawCurs(curs.row,curs.col);
				drawCursSwap(curs.row,curs.col);
			}
			else //dark to light
			{
				drawCursSwap(curs.row,curs.col);
				drawCurs(curs.row,curs.col);
			}
		}
		drwRtSec(mode,score,OSTimerCurVal(USER_TIMER),moves);
		if (key==KEY_ESC) break;
	}
	OSFreeTimer(USER_TIMER);
	for (i = 0; i < 10; i++)
  free (field[i]);
  free(field);
  //game over
  ClrScrAll();
  for(i=0;i<2;i++)
  {
  	GraySetAMSPlane(i);
  	FontSetSys(F_8x10);
	  DrawStr(40,5,"GAME OVER!",A_NORMAL);
	  FontSetSys(F_6x8);
	  DrawStr(20,30,"Your score:",A_NORMAL);
	  DrawInt(20,38,score,A_NORMAL);
	  DrawStr(20,50,"High score:",A_NORMAL);
	  DrawInt(20,58,mode?hsPuzzle[0]:hsTimed[0],A_NORMAL);
	  if (score>=(mode?hsPuzzle[0]:hsTimed[0]))
	  DrawStr(20,70,"New high score!",A_NORMAL);
  }
  //put in highscore
  if (mode==TIMED)
  {
  	hsTimed[10]=score;
  	qsort(hsTimed,11,sizeof(long),(compare_t)cmp);
  }
  else
  {
  	hsPuzzle[10]=score;
  	qsort(hsPuzzle,11,sizeof(long),(compare_t)cmp);
  }
  while (ngetchx()!=KEY_ENTER);
}

void _main(void)
{
	SetCurClip (&wholeScn);
	randomize();
	ClrScr();
	if (!GrayOn()) return;
	unsigned long pic1[2][32]={{0x000FFFFF,0x00180000,0x00380000,0x00780000,0x00F80000,0x01F80000,0x03F80000,0x07F80000,0x0FF80000,0x1FF80000,0x3FF80000,0x7FF80000,0xFFF80000,0xFFF80000,0xFFF80000,0xFFF80000,0xFFF80000,0xFFF80000,0xFFF80000,0xFFF80000,0x7FF80000,0x3FF80000,0x1FF80000,0x0FF80000,0x07F80000,0x03F80000,0x01F80000,0x00F80000,0x00780000,0x00380000,0x00180000,0x000FFFFF},{0xFFFFF000,0x00001800,0x00001C00,0x00001E00,0x00001F00,0x00001F80,0x00001FC0,0x00001FE0,0x00001FF0,0x00001FF8,0x00001FFC,0x00001FFE,0x00001FFF,0x00001FFF,0x00001FFF,0x00001FFF,0x00001FFF,0x00001FFF,0x00001FFF,0x00001FFF,0x00001FFE,0x00001FFC,0x00001FF8,0x00001FF0,0x00001FE0,0x00001FC0,0x00001F80,0x00001F00,0x00001E00,0x00001C00,0x00001800,0xFFFFF000}};
	unsigned long pic2[2][32]={{0x000FFFFF,0x001FFFFF,0x002FFFFF,0x004F3FFF,0x008F1FFF,0x010F5FFF,0x020F4FFF,0x040EEFFF,0x080EF7FF,0x100E07FF,0x200EFBFF,0x400EFBFF,0x800EFDFF,0x800EFFFF,0x800FFFFF,0x800FFFFF,0x800FFFFF,0x800F7FAE,0x800EBFAE,0x800E2222,0x400EAEAE,0x200EAE22,0x100FFFFF,0x080FFFFE,0x040FFFFE,0x020FFFFE,0x010FFFFE,0x008FFFFE,0x004FFFFE,0x002FFFFF,0x001FFFFF,0x000FFFFF},{0xFFFFF000,0xFFFFF800,0xFFFFF400,0xF87FF200,0xFB1FF100,0xFBDFF080,0xFBDFF040,0xFB9FF020,0xF83FF010,0xFBFFF008,0xFBFFF004,0xFBFFF002,0xFBFFF001,0xFBFFF001,0xFBFFF001,0xFBFFF001,0xFFFFF001,0xF7FFF001,0xFBFFF001,0x33FFF001,0xABFFF002,0xB3FFF004,0xFFFFF008,0x7FFBF010,0xBFFA3020,0x7FFAF040,0xE2237080,0xEEABB100,0xEE223200,0xFFFFF400,0xFFFFF800,0xFFFFF000}};
	Sprite32(48,34,32,pic1[0],GrayGetPlane(0),SPRT_XOR);
	Sprite32(80,34,32,pic1[1],GrayGetPlane(0),SPRT_XOR);
	Sprite32(48,34,32,pic2[0],GrayGetPlane(1),SPRT_XOR);
	Sprite32(80,34,32,pic2[1],GrayGetPlane(1),SPRT_XOR);
	ngetchx();
	ClrScrAll();
	static MULTI_LINE runespelLogo={27,{{1,10,3,10,34},{1,10,3,22,8},{1,22,8,10,14},{1,10,14,24,19},
	{1,29,30,29,5},{1,29,5,40,8},{1,40,8,40,31},
	{1,46,32,46,12},{1,46,12,61,18},{1,61,18,61,4},
	{1,64,33,64,4},{1,64,4,71,10},{1,71,10,78,4},{1,78,4,78,33},
	{1,81,4,81,19},{1,81,19,93,11},{1,93,11,93,28},
	{1,100,14,107,7},{1,107,7,107,35},
	{1,114,31,114,5},{1,114,5,120,10},{1,120,10,126,5},{1,126,5,126,31},
	{1,133,4,133,31},{1,128,17,130,12},{1,137,21,130,12},{1,137,21,139,16}}};
	
	int choice=0;//play, help, about, highscores, grayscale, quit)
	int key;
	while (choice!=5)
	{
		choice=0;
		int i;
		ClrScrAll();
		for(i=0;i<2;i++)
		{
			GraySetAMSPlane(1-i);
			DrawMultiLines(i,i,&runespelLogo);
			FontSetSys(F_8x10);
			DrawStr(10+i,39+i,"Play",A_NORMAL);
			DrawStr(10+i,49+i,"Instructions",A_NORMAL);
			DrawStr(10+i,59+i,"About",A_NORMAL);
			DrawStr(10+i,69+i,"Highscores",A_NORMAL);
			DrawStr(10+i,79+i,"Adjust Grayscale",A_NORMAL);
			DrawStr(10+i,89+i,"Quit",A_NORMAL);
		}
		drawRBlk(0,39,1);
		while ((key=ngetchx())!=KEY_ENTER)
		{
			drawRBlk(0,39+10*choice,1);
			if (key==KEY_UP) choice--;
			if (key==KEY_DOWN) choice++;
			if (choice==-1) choice=5;
			if (choice==6) choice=0;
			drawRBlk(0,39+10*choice,1);
		}
		if (choice==0)//Play
		{
			ClrScrAll();
			for(i=0;i<2;i++)
			{
				GraySetAMSPlane(1-i);
				DrawMultiLines(i,i,&runespelLogo);
				FontSetSys(F_8x10);
				DrawStr(10+i,39+i,"Timed",A_NORMAL);
				DrawStr(10+i,49+i,"Puzzle",A_NORMAL);
			}
			int mode=0;
			drawRBlk(0,39+10*mode,6);
			while ((key=ngetchx())!=KEY_ENTER)
			{
				drawRBlk(0,39+10*mode,6);
				if (key==KEY_UP) mode--;
				if (key==KEY_DOWN) mode++;
				if (mode==-1) mode=1;
				if (mode==2) mode=0;
				drawRBlk(0,39+10*mode,6);
				if (key==KEY_ESC) break;
			}
			if (key!=KEY_ESC) startSession(mode);
		}
		if (choice==1)//Help
		{
			ClrScrAll();
			for(i=0;i<2;i++)
			{
				GraySetAMSPlane(1-i);
				DrawMultiLines(i,i,&runespelLogo);
				MoveTo(40,40);
				FontSetSys(F_4x6);
				printf("How to play Runespel\nUse arrow keys to move; ENTER/F1+an arrow key will swap two symbols. Getting three or  more identical symbols in a row orthogonally will remove them and award you points.\nIn Timed mode, you have a limited amount of   time. In Puzzle mode, you have a limited\namount of moves.");
			}
			ngetchx();
		}
		if (choice==2)//about
		{
			ClrScrAll();
			for(i=0;i<2;i++)
			{
				GraySetAMSPlane(1-i);
				FontSetSys(F_8x10);
				MoveTo(0,40);
				printf("RUNESPEL\n");
				FontSetSys(F_6x8);
				printf("An ARDTHA PRODUCTION program\nProgramming and Graphics by blue_bear_94\nThis program was written for a contest against Nspire_Lover.\nWritten and compiled with TIGCC C compiler (http://tigcc.ticalc.org/)");
			}
			ngetchx();
		}
		if (choice==3)//highscores
		{
			ClrScrAll();
			for(i=0;i<2;i++)
			{
				GraySetAMSPlane(1-i);
				DrawMultiLines(i,i,&runespelLogo);
				FontSetSys(F_8x10);
				DrawStr(10+i,39+i,"Timed",A_NORMAL);
				DrawStr(10+i,49+i,"Puzzle",A_NORMAL);
			}
			int mode=0;
			drawRBlk(0,39+10*mode,5);
			while ((key=ngetchx())!=KEY_ENTER)
			{
				drawRBlk(0,39+10*mode,5);
				if (key==KEY_UP) mode--;
				if (key==KEY_DOWN) mode++;
				if (mode==-1) mode=1;
				if (mode==2) mode=0;
				drawRBlk(0,39+10*mode,5);
				if (key==KEY_ESC) break;
			}
			if (key!=KEY_ESC)
			{
				long* show=mode?hsPuzzle:hsTimed;
				ClrScrAll();
				for(i=0;i<2;i++)
				{
					GraySetAMSPlane(1-i);
					DrawMultiLines(i,i,&runespelLogo);
					FontSetSys(F_8x10);
					DrawStr(i,39+i,"Highscores",A_NORMAL);
					FontSetSys(F_4x6);
					int j;
					char buff[12];
					for(j=0;j<10;j++)
					{
						sprintf(buff,"%02d:%08ld",j+1,show[j]);
						DrawStr(100,40+6*j,buff,A_NORMAL);
					}
				}
				ngetchx();
			}
		}
		if (choice==4)//adjust grayscale
		{
			ClrScrAll();
			SCR_RECT* temp=&(SCR_RECT){{40,0,79,99}};
			GraySetAMSPlane(LIGHT_PLANE);
			ScrRectFill(temp,&wholeScn,A_NORMAL);
			temp->xy.x0+=80;
			temp->xy.x1+=80;
			ScrRectFill(temp,&wholeScn,A_NORMAL);
			GraySetAMSPlane(DARK_PLANE);
			ScrRectFill(temp,&wholeScn,A_NORMAL);
			temp->xy.x0-=40;
			temp->xy.x1-=40;
			ScrRectFill(temp,&wholeScn,A_NORMAL);
			FontSetSys(F_4x6);
			for(i=0;i<2;i++)
			{
				GraySetAMSPlane(i);
				DrawStr(0,0,"Use \x17 and \x18 to adjust",A_NORMAL);
				DrawStr(0,6,"grayscale quality (ENTER",A_NORMAL);
				DrawStr(0,12,"/ESC to exit)",A_NORMAL);
			}
			int value=0;
			GrayAdjust(value);
			while ((key=ngetchx())!=KEY_ENTER && key!=KEY_ESC)
			{
	      if (key== KEY_UP && value < 127)
	        value++;
	      if (key== KEY_DOWN && value > (TI89 ? -28 : 0))
	        value--;
	      GrayAdjust(value);
			}
			temp->xy.x0-=40;
			temp->xy.x1-=40;
		}
	}
	GrayOff();
}
